from functools import partialmethod
from typing import Optional, Dict, List, AsyncGenerator

from .base import CallbackFnType, InputType
from ..helper import run_async
from ..types.request import Response


class PostMixin:
    """The Post Mixin class for Client and Flow"""

    def post(
        self,
        on: str,
        inputs: Optional[InputType] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        parameters: Optional[Dict] = None,
        target_peapod: Optional[str] = None,
        request_size: int = 100,
        show_progress: bool = False,
        continue_on_error: bool = False,
        return_results: bool = False,
        **kwargs,
    ) -> Optional[List[Response]]:
        """Post a general data request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document id.
        :param on: the endpoint is used for identifying the user-defined ``request_type``, labeled by ``@requests(on='/abc')``
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param parameters: the kwargs that will be sent to the executor
        :param target_peapod: a regex string represent the certain peas/pods request targeted
        :param request_size: the number of Documents per request. <=0 means all inputs in one request.
        :param show_progress: if set, client will show a progress bar on receiving every request.
        :param continue_on_error: if set, a Request that causes callback error will be logged only without blocking the further requests.
        :param return_results: if set, the results of all Requests will be returned as a list. This is useful when one wants process Responses in bulk instead of using callback.
        :param kwargs: additional parameters
        :return: None or list of Response
        """

        async def _get_results(*args, **kwargs):
            result = []
            c = self.client
            c.show_progress = show_progress
            c.continue_on_error = continue_on_error
            async for resp in c._get_results(*args, **kwargs):
                if return_results:
                    result.append(resp)

            if return_results:
                return result

        return run_async(
            _get_results,
            inputs=inputs,
            on_done=on_done,
            on_error=on_error,
            on_always=on_always,
            exec_endpoint=on,
            target_peapod=target_peapod,
            parameters=parameters,
            request_size=request_size,
            **kwargs,
        )

    # ONLY CRUD, for other request please use `.post`
    index = partialmethod(post, '/index')
    search = partialmethod(post, '/search')
    update = partialmethod(post, '/update')
    delete = partialmethod(post, '/delete')


class AsyncPostMixin:
    """The Async Post Mixin class for AsyncClient and AsyncFlow"""

    async def post(
        self,
        on: str,
        inputs: Optional[InputType] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        parameters: Optional[Dict] = None,
        target_peapod: Optional[str] = None,
        request_size: int = 100,
        show_progress: bool = False,
        continue_on_error: bool = False,
        **kwargs,
    ) -> AsyncGenerator[None, Response]:
        """Post a general data request to the Flow.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document id.
        :param on: the endpoint is used for identifying the user-defined ``request_type``, labeled by ``@requests(on='/abc')``
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param parameters: the kwargs that will be sent to the executor
        :param target_peapod: a regex string represent the certain peas/pods request targeted
        :param request_size: the number of Documents per request. <=0 means all inputs in one request.
        :param show_progress: if set, client will show a progress bar on receiving every request.
        :param continue_on_error: if set, a Request that causes callback error will be logged only without blocking the further requests.
        :param kwargs: additional parameters
        :yield: Response object
        """
        c = self.client
        c.show_progress = show_progress
        c.continue_on_error = continue_on_error
        async for r in c._get_results(
            inputs=inputs,
            on_done=on_done,
            on_error=on_error,
            on_always=on_always,
            exec_endpoint=on,
            target_peapod=target_peapod,
            parameters=parameters,
            request_size=request_size,
            **kwargs,
        ):
            yield r

    # ONLY CRUD, for other request please use `.post`
    index = partialmethod(post, '/index')
    search = partialmethod(post, '/search')
    update = partialmethod(post, '/update')
    delete = partialmethod(post, '/delete')



schema_metas = {
    'Jina::Metas': {
        'description': 'The meta config of the Executor',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'name': {
                'description': 'The name of the executor.',
                'type': 'string',
                'default': None,
            },
            'description': {
                'description': 'The description of this executor. It will be used in automatics docs UI',
                'type': 'string',
                'default': None,
            },
            'py_modules': {
                'type': 'array',
                'default': None,
                'items': {'type': 'string', 'minItems': 1, 'uniqueItems': True},
                'description': '''
The external python module paths. it is useful when you want to load external python modules
using :func:`BaseExecutor.load_config` from a YAML file. If a relative path is given then the root path is set to
the path of the current YAML file.

Example of ``py_module`` usage:

1. This is a valid structure and it is RECOMMENDED:
    - "my_cust_module" is a python module
    - all core logic of your customized executor goes to ``__init__.py``
    - to import ``foo.py``, you can use relative import, e.g. ``from .foo import bar``
    - ``helper.py`` needs to be put BEFORE `__init__.py` in YAML ``py_modules``

This is also the structure given by ``jina hub new`` CLI.

    .. highlight:: text
    .. code-block:: text

         my_cust_module
           |- __init__.py
           |- helper.py
           |- config.yml
                |- py_modules
                       |- helper.py
                       |- __init__.py

2. This is a valid structure but not recommended:
    - "my_cust_module" is not a python module (lack of __init__.py under the root)
    - to import ``foo.py``, you must to use ``from jinahub.foo import bar``
    - ``jinahub`` is a common namespace for all plugin-modules, not changeable.
    - ``helper.py`` needs to be put BEFORE `my_cust.py` in YAML ``py_modules``

    .. highlight:: text
    .. code-block:: text

         my_cust_module
           |- my_cust.py
           |- helper.py
           |- config.yml
                |- py_modules
                       |- helper.py
                       |- my_cust.py                
                ''',
            },
        },
    }
}



import argparse
import asyncio
from abc import ABC
from typing import AsyncGenerator

from ....helper import typename
from ....logging.logger import JinaLogger
from ....types.message import Message

__all__ = ['PrefetchCaller', 'PrefetchMixin']

if False:
    from ...zmq import AsyncZmqlet


class PrefetchMixin(ABC):
    """JinaRPCServicer """

    async def Call(self, request_iterator, *args) -> AsyncGenerator[None, Message]:
        """
        Async call to receive Requests and build them into Messages.

        :param request_iterator: iterator of requests.
        :param args: additional arguments
        :yield: message
        """
        self.args: argparse.Namespace
        self.zmqlet: 'AsyncZmqlet'
        self.logger: JinaLogger

        async def prefetch_req(num_req, fetch_to):
            """
            Fetch and send request.

            :param num_req: number of requests
            :param fetch_to: the task list storing requests
            :return: False if append task to :param:`fetch_to` else False
            """
            for _ in range(num_req):
                try:
                    if hasattr(request_iterator, '__anext__'):
                        next_request = await request_iterator.__anext__()
                    elif hasattr(request_iterator, '__next__'):
                        next_request = next(request_iterator)
                    else:
                        raise TypeError(
                            f'{typename(request_iterator)} does not have `__anext__` or `__next__`'
                        )
                    asyncio.create_task(
                        self.zmqlet.send_message(
                            Message(None, next_request, 'gateway', **vars(self.args))
                        )
                    )
                    fetch_to.append(
                        asyncio.create_task(
                            self.zmqlet.recv_message(callback=lambda x: x.response)
                        )
                    )
                except (StopIteration, StopAsyncIteration):
                    return True
            return False

        prefetch_task = []
        is_req_empty = await prefetch_req(self.args.prefetch, prefetch_task)
        if is_req_empty and not prefetch_task:
            self.logger.error(
                'receive an empty stream from the client! '
                'please check your client\'s inputs, '
                'you can use "Client.check_input(inputs)"'
            )
            return

        # the total num requests < self.args.prefetch
        if is_req_empty:
            for r in asyncio.as_completed(prefetch_task):
                yield await r
        else:
            # if there are left over (`else` clause above is unnecessary for code but for better readability)
            onrecv_task = []
            # the following code "interleaves" prefetch_task and onrecv_task, when one dries, it switches to the other
            while prefetch_task:
                self.logger.debug(
                    f'send: {self.zmqlet.msg_sent} '
                    f'recv: {self.zmqlet.msg_recv} '
                    f'pending: {self.zmqlet.msg_sent - self.zmqlet.msg_recv}'
                )
                onrecv_task.clear()
                for r in asyncio.as_completed(prefetch_task):
                    yield await r
                    if not is_req_empty:
                        is_req_empty = await prefetch_req(
                            self.args.prefetch_on_recv, onrecv_task
                        )

                # this list dries, clear it and feed it with on_recv_task
                prefetch_task.clear()
                prefetch_task = [j for j in onrecv_task]


class PrefetchCaller(PrefetchMixin):
    """An async zmq request sender to be used in the Gateway"""

    def __init__(self, args: argparse.Namespace, zmqlet: 'AsyncZmqlet'):
        """

        :param args: args from CLI
        :param zmqlet: zeromq object
        """
        super().__init__()
        self.args = args
        self.zmqlet = zmqlet
        self.name = args.name or self.__class__.__name__
        self.logger = JinaLogger(self.name, **vars(args))



"""Module wrapping interactions with the local executor packages."""

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Tuple, Optional

from .helper import unpack_package

_hub_root = Path(
    os.environ.get('JINA_HUB_ROOT', Path.home().joinpath('.jina', 'hub-packages'))
)
_hub_root.mkdir(parents=True, exist_ok=True)


def get_dist_path(uuid: str, tag: str) -> Tuple['Path', 'Path']:
    """Get the package path according ID and TAG
    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: package and its dist-info path
    """
    pkg_path = _hub_root / uuid
    pkg_dist_path = _hub_root / f'{uuid}-{tag}.dist-info'
    return pkg_path, pkg_dist_path


def _install_requirements(requirements_file: 'Path'):
    """Install modules included in requirments file

    :param requirements_file: the requirements.txt file
    """
    subprocess.check_call(
        [sys.executable, '-m', 'pip', 'install', '-r', f'{requirements_file}']
    )


def install_local(
    zip_package: 'Path',
    uuid: str,
    tag: str,
    force: Optional[bool] = False,
    install_deps: Optional[bool] = False,
):
    """Install the package in zip format to the Jina Hub root.

    :param zip_package: the path of the zip file
    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :param force: if set, overwrites the package
    :param install_deps: if set, install dependencies
    """

    pkg_path, pkg_dist_path = get_dist_path(uuid, tag)

    if pkg_dist_path.exists() and not force:
        return

    # clean existed dist-info
    for dist in _hub_root.glob(f'{uuid}-*.dist-info'):
        shutil.rmtree(dist)
    if pkg_path.exists():
        shutil.rmtree(pkg_path)

    try:
        # unpack the zip package to the root pkg_path
        unpack_package(zip_package, pkg_path)

        # create dist-info folder
        pkg_dist_path.mkdir(parents=False, exist_ok=True)

        # install the dependencies included in requirements.txt
        if install_deps:
            requirements_file = pkg_path / 'requirements.txt'
            if requirements_file.exists():
                _install_requirements(requirements_file)
                shutil.copyfile(requirements_file, pkg_dist_path / 'requirements.txt')

        manifest_path = pkg_path / 'manifest.yml'
        if manifest_path.exists():
            shutil.copyfile(manifest_path, pkg_dist_path / 'manifest.yml')

    except Exception as ex:
        # clean pkg_path, pkg_dist_path
        shutil.rmtree(pkg_path)
        shutil.rmtree(pkg_dist_path)
        raise ex


def uninstall_local(uuid: str):
    """Uninstall the executor package.

    :param uuid: the UUID of the executor
    """
    pkg_path, _ = get_dist_path(uuid, None)
    for dist in _hub_root.glob(f'{uuid}-*.dist-info'):
        shutil.rmtree(dist)
    if pkg_path.exists():
        shutil.rmtree(pkg_path)


def list_local():
    """List the locally-available executor packages.

    :return: the list of local executors (if found)
    """
    result = []
    for dist_name in _hub_root.glob(r'*-v*.dist-info'):
        result.append(dist_name)

    return result


def resolve_local(uuid: str, tag: Optional[str] = None) -> Optional['Path']:
    """Return the path of the executor if available.

    :param uuid: the UUID of executor
    :param tag: the TAG of executor
    :return: the path of the executor package
    """
    pkg_path = _hub_root / uuid
    pkg_dist_path = _hub_root / f'{uuid}-{tag}.dist-info'
    if not pkg_path.exists() or (tag and not pkg_dist_path.exists()):
        raise FileNotFoundError(f'{pkg_path} doe not exist')
    else:
        return pkg_path


def exist_local(uuid: str, tag: str = None) -> bool:
    """Check whether the executor exists in local

    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: True if existed, else False
    """
    try:
        resolve_local(uuid, tag=tag)
        return True
    except FileNotFoundError:
        return False



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import serializer as jina__pb2


class JinaRPCStub(object):
    """*
    jina gRPC service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Call = channel.stream_stream(
                '/jina.JinaRPC/Call',
                request_serializer=jina__pb2.RequestProto.SerializeToString,
                response_deserializer=jina__pb2.RequestProto.FromString,
                )


class JinaRPCServicer(object):
    """*
    jina gRPC service.
    """

    def Call(self, request_iterator, context):
        """Pass in a Request and a filled Request with matches will be returned.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JinaRPCServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Call': grpc.stream_stream_rpc_method_handler(
                    servicer.Call,
                    request_deserializer=jina__pb2.RequestProto.FromString,
                    response_serializer=jina__pb2.RequestProto.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'jina.JinaRPC', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class JinaRPC(object):
    """*
    jina gRPC service.
    """

    @staticmethod
    def Call(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/jina.JinaRPC/Call',
            jina__pb2.RequestProto.SerializeToString,
            jina__pb2.RequestProto.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from ..base import set_base_parser
from ..helper import _chf


def set_hub_push_parser(parser=None):
    """Set the parser for the hub push
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from .push import mixin_hub_push_parser

    mixin_hub_push_parser(parser)
    return parser


def set_hub_pull_parser(parser=None):
    """Set the parser for the hub pull
    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from .pull import mixin_hub_pull_parser

    mixin_hub_pull_parser(parser)
    return parser


def set_hub_parser(parser=None):
    """Set the parser for the hub
    :param parser: the parser configure

    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    spp = parser.add_subparsers(
        dest='hub',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get detailed information about each sub-command',
        required=True,
    )

    set_hub_push_parser(
        spp.add_parser(
            'push',
            help='push an executor package to the Jina hub',
            description='Push an executor package to the Jina hub',
            formatter_class=_chf,
        )
    )

    set_hub_pull_parser(
        spp.add_parser(
            'pull',
            help='download an executor package/image from the Jina hub',
            description='Download an executor package/image from the Jina hub',
            formatter_class=_chf,
        )
    )

    return parser



"""Module for wrapping Jina Hub API calls."""

import argparse
import hashlib
import json
import os
from collections import namedtuple
from pathlib import Path
from typing import Optional, Dict
from urllib.parse import urlencode

from .helper import archive_package, download_with_resume, parse_hub_uri, get_hubble_url
from .hubapi import install_local, exist_local
from ..excepts import HubDownloadError
from ..helper import colored, get_full_version, get_readable_size, ArgNamespace
from ..importer import ImportExtensions
from ..logging.logger import JinaLogger
from ..logging.profile import TimeContext
from ..parsers.hubble import set_hub_parser

HubExecutor = namedtuple(
    'HubExecutor',
    ['uuid', 'alias', 'tag', 'visibility', 'image_name', 'archive_url', 'md5sum'],
)


class HubIO:
    """:class:`HubIO` provides the way to interact with Jina Hub registry.
    You can use it with CLI to package a directory into a Jina Hub and publish it to the world.
    Examples:
        - :command:`jina hub push my_executor/` to push the executor package to Jina Hub
        - :command:`jina hub pull UUID8` to download the executor identified by UUID8

    To create a :class:`HubIO` object, simply:

        .. highlight:: python
        .. code-block:: python
            hubio = HubIO(args)

    :param args: arguments
    """

    def __init__(self, args: Optional[argparse.Namespace] = None, **kwargs):
        if args and isinstance(args, argparse.Namespace):
            self.args = args
        else:
            self.args = ArgNamespace.kwargs2namespace(kwargs, set_hub_parser())
        self.logger = JinaLogger(self.__class__.__name__, **vars(args))

        self._load_docker_client()

    def _load_docker_client(self):
        with ImportExtensions(
            required=True,
            help_text='missing "docker" dependency, please do pip install "jina[docker]"',
        ):
            import docker
            from docker import DockerClient

            self._client: DockerClient = docker.from_env()

    @staticmethod
    def _get_request_header() -> Dict:
        """Return the header of request.

        :return: request header
        """
        metas, envs = get_full_version()

        header = {
            **{f'jinameta-{k}': str(v) for k, v in metas.items()},
            **envs,
        }
        return header

    def push(self) -> None:
        """Push the executor pacakge to Jina Hub."""

        pkg_path = Path(self.args.path)
        if not pkg_path.exists():
            self.logger.critical(f'`{self.args.path}` is not a valid path!')
            exit(1)

        request_headers = self._get_request_header()

        try:
            # archive the executor package
            with TimeContext(f'Packaging {self.args.path}', self.logger):
                md5_hash = hashlib.md5()
                bytesio = archive_package(pkg_path)
                content = bytesio.getvalue()
                md5_hash.update(content)
                md5_digest = md5_hash.hexdigest()

            # upload the archived package
            form_data = {
                'public': 'True' if getattr(self.args, 'public', None) else 'False',
                'private': 'True' if getattr(self.args, 'private', None) else 'False',
                'md5sum': md5_digest,
            }
            if self.args.force:
                form_data['force'] = self.args.force
            if self.args.secret:
                form_data['secret'] = self.args.secret

            method = 'put' if self.args.force else 'post'

            hubble_url = get_hubble_url()
            # upload the archived executor to Jina Hub
            with TimeContext(
                f'Pushing to {hubble_url} ({method.upper()})',
                self.logger,
            ):
                from .helper import upload_file

                resp = upload_file(
                    hubble_url,
                    'filename',
                    content,
                    dict_data=form_data,
                    headers=request_headers,
                    stream=True,
                    method=method,
                )

                result = None
                for stream_line in resp.iter_lines():
                    stream_msg = json.loads(stream_line)
                    if 'stream' in stream_msg:
                        self.logger.info(stream_msg['stream'])
                    elif 'result' in stream_msg:
                        result = stream_msg['result']
                        break

                if result is None:
                    raise Exception('Unknown Error')

                if 200 <= result['statusCode'] < 300:
                    # TODO: only support single executor now
                    data = result.get('data', None)
                    if not data:
                        raise Exception(result.get('message', 'Unknown Error'))

                    image = data['executors'][0]

                    uuid8 = image['id']
                    secret = image['secret']
                    visibility = image['visibility']

                    info_table = [
                        f'\t🔑 ID:\t\t' + colored(f'{uuid8}', 'cyan'),
                        f'\t🔒 Secret:\t'
                        + colored(
                            f'{secret}',
                            'cyan',
                        )
                        + colored(
                            ' (👈 Please store this secret carefully, it wont show up again)',
                            'red',
                        ),
                        f'\t👀 Visibility:\t' + colored(f'{visibility}', 'cyan'),
                    ]

                    if 'alias' in image:
                        info_table.append(
                            f'\t📛 Alias:\t' + colored(image['alias'], 'cyan')
                        )

                    self.logger.info('\n' + '\n'.join(info_table))

                    usage = (
                        f'jinahub://{uuid8}'
                        if visibility == 'public'
                        else f'jinahub://{uuid8}:{secret}'
                    )

                    self.logger.info(
                        f'You can use it via `uses={usage}` in the Flow/CLI.'
                    )
                elif result['message']:
                    raise Exception(result['message'])
                elif resp.text:
                    # NOTE: sometimes resp.text returns empty
                    raise Exception(resp.text)
                else:
                    resp.raise_for_status()
        except Exception as e:  # IO related errors
            self.logger.error(
                f'Error while pushing `{self.args.path}` with session_id={request_headers["jinameta-session-id"]}: '
                f'\n{e!r}'
            )

    @staticmethod
    def fetch(
        name: str,
        tag: Optional[str] = None,
        secret: Optional[str] = None,
    ) -> HubExecutor:
        """Fetch the executor meta info from Jina Hub.
        :param name: the UUID/Alias of the executor
        :param tag: the version tag of the executor
        :param secret: the access secret of the executor
        :return: meta of executor
        """

        with ImportExtensions(required=True):
            import requests

        pull_url = get_hubble_url() + f'/{name}/?'
        path_params = {}
        if secret:
            path_params['secret'] = secret
        if tag:
            path_params['tag'] = tag

        request_headers = HubIO._get_request_header()

        pull_url += urlencode(path_params)
        resp = requests.get(pull_url, headers=request_headers)
        if resp.status_code != 200:
            if resp.text:
                raise Exception(resp.text)
            resp.raise_for_status()

        resp = resp.json()

        result = HubExecutor(
            resp['id'],
            resp.get('alias', None),
            resp['tag'],
            resp['visibility'],
            resp['image'],
            resp['package']['download'],
            resp['package']['md5'],
        )

        return result

    def pull(self) -> None:
        """Pull the executor package from Jina Hub."""
        cached_zip_filepath = None
        try:
            scheme, name, tag, secret = parse_hub_uri(self.args.uri)

            executor = HubIO.fetch(name, tag=tag, secret=secret)

            if not tag:
                tag = executor.tag

            uuid = executor.uuid
            image_name = executor.image_name
            archive_url = executor.archive_url
            md5sum = executor.md5sum

            if scheme == 'jinahub+docker':
                # pull the Docker image
                with TimeContext(f'pulling {image_name}', self.logger):
                    image = self._client.images.pull(image_name)
                if isinstance(image, list):
                    image = image[0]
                image_tag = image.tags[0] if image.tags else ''
                self.logger.success(
                    f'🎉 pulled {image_tag} ({image.short_id}) uncompressed size: {get_readable_size(image.attrs["Size"])}'
                )
                return
            if exist_local(uuid, tag):
                self.logger.debug(
                    f'The executor `{self.args.uri}` has already been downloaded.'
                )
                return
            # download the package
            with TimeContext(f'downloading {self.args.uri}', self.logger):
                cache_dir = Path(
                    os.environ.get(
                        'JINA_HUB_CACHE_DIR', Path.home().joinpath('.cache', 'jina')
                    )
                )
                cache_dir.mkdir(parents=True, exist_ok=True)
                cached_zip_filename = f'{uuid}-{md5sum}.zip'
                cached_zip_filepath = download_with_resume(
                    archive_url,
                    cache_dir,
                    cached_zip_filename,
                    md5sum=md5sum,
                )

            with TimeContext(f'unpacking {self.args.uri}', self.logger):
                try:
                    install_local(
                        cached_zip_filepath,
                        uuid,
                        tag,
                        install_deps=self.args.install_requirements,
                    )
                except Exception as ex:
                    raise HubDownloadError(str(ex))

        except Exception as e:
            self.logger.error(
                f'Error when pulling the executor `{self.args.uri}`: {e!r}'
            )
        finally:
            # delete downloaded zip package if existed
            if cached_zip_filepath is not None:
                cached_zip_filepath.unlink()



import argparse
import re
import time
from collections import defaultdict
from typing import Dict, List

import zmq

from .base import ZMQRuntime
from ...zmq import ZmqStreamlet
from .... import __default_endpoint__
from ....enums import OnErrorStrategy, SocketType
from ....excepts import (
    NoExplicitMessage,
    ExecutorFailToLoad,
    MemoryOverHighWatermark,
    ChainedPodException,
    BadConfigSource,
    RuntimeTerminated,
    UnknownControlCommand,
)
from ....executors import BaseExecutor
from ....helper import random_identity, typename
from ....logging.profile import used_memory
from ....proto import jina_pb2
from ....types.arrays.document import DocumentArray
from ....types.message import Message
from ....types.request import Request
from ....types.routing.table import RoutingTable


class ZEDRuntime(ZMQRuntime):
    """Runtime procedure leveraging :class:`ZmqStreamlet` for Executor."""

    def __init__(self, args: argparse.Namespace, ctrl_addr: str, **kwargs):
        """Initialize private parameters and execute private loading functions.

        :param args: args from CLI
        :param ctrl_addr: control port address
        :param kwargs: extra keyword arguments
        """
        super().__init__(args, ctrl_addr, **kwargs)
        self._id = random_identity()
        self._last_active_time = time.perf_counter()

        self._request = None
        self._message = None

        # all pending messages collected so far, key is the request id
        self._pending_msgs = defaultdict(list)  # type: Dict[str, List['Message']]
        self._partial_requests = None
        self._partial_messages = None

        # idle_dealer_ids only becomes non-None when it receives IDLE ControlRequest
        self._idle_dealer_ids = set()

        self._load_zmqstreamlet()
        self._load_plugins()
        self._load_executor()

    def run_forever(self):
        """Start the `ZmqStreamlet`."""
        self._zmqstreamlet.start(self._msg_callback)

    def teardown(self):
        """Close the `ZmqStreamlet` and `Executor`."""
        self._zmqstreamlet.close()
        self._executor.close()
        super().teardown()

    #: Private methods required by :meth:`setup`

    def _load_zmqstreamlet(self):
        """Load ZMQStreamlet to this runtime."""
        # important: fix zmqstreamlet ctrl address to replace the the ctrl address generated in the main
        # process/thread
        self._zmqstreamlet = ZmqStreamlet(
            args=self.args,
            logger=self.logger,
            ctrl_addr=self.ctrl_addr,
            ready_event=self.is_ready_event,
        )

    def _load_executor(self):
        """Load the executor to this runtime, specified by ``uses`` CLI argument."""
        try:
            self._executor = BaseExecutor.load_config(
                self.args.uses,
                override_with=self.args.override_with,
                override_metas=self.args.override_metas,
                runtime_args=vars(self.args),
            )
        except BadConfigSource as ex:
            self.logger.error(
                f'fail to load config from {self.args.uses}, if you are using docker image for --uses, '
                f'please use "docker://YOUR_IMAGE_NAME"'
            )
            raise ExecutorFailToLoad from ex
        except FileNotFoundError as ex:
            self.logger.error(f'fail to load file dependency')
            raise ExecutorFailToLoad from ex
        except Exception as ex:
            self.logger.critical(f'can not load the executor from {self.args.uses}')
            raise ExecutorFailToLoad from ex

    def _load_plugins(self):
        """Load the plugins if needed necessary to load executors."""
        if self.args.py_modules:
            from ....importer import PathImporter

            PathImporter.add_modules(*self.args.py_modules)

    #: Private methods required by :meth:`teardown`
    def _check_memory_watermark(self):
        """Check the memory watermark."""
        if used_memory() > self.args.memory_hwm > 0:
            raise MemoryOverHighWatermark

    #: Private methods required by run_forever
    def _pre_hook(self, msg: 'Message') -> 'ZEDRuntime':
        """
        Pre-hook function, what to do after first receiving the message.

        :param msg: received message
        :return: `ZEDRuntime`
        """
        msg.add_route(self.name, self._id)
        self._request = msg.request
        self._message = msg

        part_str = ''
        if self.expect_parts > 1:
            req_id = msg.envelope.request_id
            self._pending_msgs[req_id].append(msg)
            self._partial_messages = self._pending_msgs[req_id]
            self._partial_requests = [v.request for v in self._partial_messages]
            part_str = f'({len(self.partial_requests)}/{self.expect_parts} parts)'

        info_msg = f'recv {msg.envelope.request_type} '
        if self.request_type == 'DataRequest':
            info_msg += f'({self.envelope.header.exec_endpoint}) '
        elif self.request_type == 'ControlRequest':
            info_msg += f'({self.request.command}) '
        info_msg += f'{part_str} from {msg.colored_route}'
        self.logger.debug(info_msg)

        if self.expect_parts > 1 and self.expect_parts > len(self.partial_requests):
            # NOTE: reduce priority is higher than chain exception
            # otherwise a reducer will lose its function when earlier pods raise exception
            raise NoExplicitMessage

        if self.request_type == 'ControlRequest':
            self._handle_control_req()

        if (
            msg.envelope.status.code == jina_pb2.StatusProto.ERROR
            and self.args.on_error_strategy >= OnErrorStrategy.SKIP_HANDLE
        ):
            raise ChainedPodException

        return self

    def _post_hook(self, msg: 'Message') -> 'ZEDRuntime':
        """
        Post-hook function, what to do before handing out the message.

        :param msg: received message
        :return: `ZEDRuntime`
        """
        # do NOT access `msg.request.*` in the _pre_hook, as it will trigger the deserialization
        # all meta information should be stored and accessed via `msg.envelope`

        self._last_active_time = time.perf_counter()
        self._check_memory_watermark()

        if self.expect_parts > 1:
            msgs = self._pending_msgs.pop(msg.envelope.request_id)
            msg.merge_envelope_from(msgs)

        msg.update_timestamp()
        return self

    @staticmethod
    def _parse_params(parameters: Dict, executor_name: str):
        parsed_params = parameters
        specific_parameters = parameters.get(executor_name, None)
        if specific_parameters:
            parsed_params.update(**specific_parameters)
        return parsed_params

    def _handle(self) -> 'ZEDRuntime':
        """Register the current message to this pea, so that all message-related properties are up-to-date, including
        :attr:`request`, :attr:`prev_requests`, :attr:`message`, :attr:`prev_messages`. And then call the executor to handle
        this message if its envelope's  status is not ERROR, else skip handling of message.

        .. note::
            Handle does not handle explicitly message because it may wait for different messages when different parts are expected
        :return: ZEDRuntime procedure.
        """

        if (
            not re.match(self.envelope.header.target_peapod, self.name)
            or self.request_type != 'DataRequest'
        ):
            return self

        # migrated from the previously RouteDriver logic
        if self._idle_dealer_ids:
            dealer_id = self._idle_dealer_ids.pop()
            self.envelope.receiver_id = dealer_id

            # when no available dealer, pause the pollin from upstream
            if not self._idle_dealer_ids:
                self._zmqstreamlet.pause_pollin()

        if (
            self.envelope.header.exec_endpoint not in self._executor.requests
            and __default_endpoint__ not in self._executor.requests
        ):
            return self

        params = self._parse_params(self.request.parameters, self._executor.metas.name)

        # executor logic
        r_docs = self._executor(
            req_endpoint=self.envelope.header.exec_endpoint,
            docs=self.docs,
            parameters=params,
            docs_matrix=self.docs_matrix,
            groundtruths=self.groundtruths,
            groundtruths_matrix=self.groundtruths_matrix,
        )

        # assigning result back to request
        # 1. Return none: do nothing
        # 2. Return nonempty and non-DocumentArray: raise error
        # 3. Return DocArray, but the memory pointer says it is the same as self.docs: do nothing
        # 4. Return DocArray and its not a shallow copy of self.docs: assign self.request.docs
        if r_docs is not None:
            if not isinstance(r_docs, DocumentArray):
                raise TypeError(
                    f'return type must be {DocumentArray!r} or None, but getting {typename(r_docs)}'
                )
            elif r_docs != self.request.docs:
                # this means the returned DocArray is a completely new one
                self.request.docs.clear()
                self.request.docs.extend(r_docs)

        return self

    def _handle_control_req(self):
        # migrated from previous ControlDriver logic
        if self.request.command == 'TERMINATE':
            self.envelope.status.code = jina_pb2.StatusProto.SUCCESS
            raise RuntimeTerminated
        elif self.request.command == 'STATUS':
            self.envelope.status.code = jina_pb2.StatusProto.READY
            self.request.parameters = vars(self.args)
        elif self.request.command == 'IDLE':
            self._idle_dealer_ids.add(self.envelope.receiver_id)
            self._zmqstreamlet.resume_pollin()
            self.logger.debug(
                f'{self.envelope.receiver_id} is idle, now I know these idle peas {self._idle_dealer_ids}'
            )
        elif self.request.command == 'CANCEL':
            if self.envelope.receiver_id in self._idle_dealer_ids:
                self._idle_dealer_ids.remove(self.envelope.receiver_id)
        elif self.request.command == 'ACTIVATE':
            self._zmqstreamlet._send_idle_to_router()
        elif self.request.command == 'DEACTIVATE':
            self._zmqstreamlet._send_cancel_to_router()
        else:
            raise UnknownControlCommand(
                f'don\'t know how to handle {self.request.command}'
            )

    def _callback(self, msg: 'Message'):
        self.is_post_hook_done = False  #: if the post_hook is called
        self._pre_hook(msg)._handle()._post_hook(msg)
        self.is_post_hook_done = True
        return msg

    def _msg_callback(self, msg: 'Message') -> None:
        """
        Callback function after receiving the message
        When nothing is returned then nothing is send out via :attr:`zmqlet.sock_out`.

        :param msg: received message
        """
        try:
            # notice how executor related exceptions are handled here
            # generally unless executor throws an OSError, the exception are caught and solved inplace
            self._zmqstreamlet.send_message(self._callback(msg))
        except RuntimeTerminated:
            # this is the proper way to end when a terminate signal is sent
            self._zmqstreamlet.send_message(msg)
            self._zmqstreamlet.close()
        except KeyboardInterrupt as kbex:
            # save executor
            self.logger.debug(f'{kbex!r} causes the breaking from the event loop')
            self._zmqstreamlet.send_message(msg)
            self._zmqstreamlet.close(flush=False)
        except (SystemError, zmq.error.ZMQError) as ex:
            # save executor
            self.logger.debug(f'{ex!r} causes the breaking from the event loop')
            self._zmqstreamlet.send_message(msg)
            self._zmqstreamlet.close()
        except MemoryOverHighWatermark:
            self.logger.critical(
                f'memory usage {used_memory()} GB is above the high-watermark: {self.args.memory_hwm} GB'
            )
        except NoExplicitMessage:
            # silent and do not propagate message anymore
            # 1. wait partial message to be finished
            # 2. dealer send a control message and no need to go on
            pass
        except (RuntimeError, Exception, ChainedPodException) as ex:
            # general runtime error and nothing serious, we simply mark the message to error and pass on
            if not self.is_post_hook_done:
                self._post_hook(msg)

            if self.args.on_error_strategy == OnErrorStrategy.THROW_EARLY:
                raise
            if isinstance(ex, ChainedPodException):
                # the error is print from previous pod, no need to show it again
                # hence just add exception and propagate further
                # please do NOT add logger.error here!
                msg.add_exception()
            else:
                msg.add_exception(ex, executor=getattr(self, '_executor'))
                self.logger.error(
                    f'{ex!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else '',
                    exc_info=not self.args.quiet_error,
                )

            self._zmqstreamlet.send_message(msg)

    #: Some class-specific properties

    @property
    def is_idle(self) -> bool:
        """
        Return ``True`` when current time is ``max_idle_time`` seconds late than the last active time

        :return: True if idle else false.
        """
        return (time.perf_counter() - self._last_active_time) > self.args.max_idle_time

    @property
    def request(self) -> 'Request':
        """
        Get the current request body inside the protobuf message

        :return: :class:`ZEDRuntime` request
        """
        return self._request

    @property
    def message(self) -> 'Message':
        """
        Get the current protobuf message to be processed

        :return: :class:`ZEDRuntime` message
        """
        return self._message

    @property
    def request_type(self) -> str:
        """
        Get the type of message being processed

        :return: request type
        """
        return self._message.envelope.request_type

    @property
    def expect_parts(self) -> int:
        """
        The expected number of partial messages before trigger :meth:`handle`

        :return: expected number of partial messages
        """
        if self.message.is_data_request:
            if self.args.socket_in == SocketType.ROUTER_BIND:
                graph = RoutingTable(self._message.envelope.routing_table)
                return graph.active_target_pod.expected_parts
            else:
                return self.args.num_part
        else:
            return 1

    @property
    def partial_requests(self) -> List['Request']:
        """
        The collected partial requests under the current ``request_id``

        :return: collected partial requests
        """
        return self._partial_requests

    @property
    def partial_messages(self) -> List['Message']:
        """
        The collected partial messages under the current ``request_id`` "
        :return: collected partial messages

        """
        return self._partial_messages

    def _get_docs(self, field: str) -> 'DocumentArray':
        if self.expect_parts > 1:
            result = DocumentArray(
                [d for r in reversed(self.partial_requests) for d in getattr(r, field)]
            )
        else:
            result = getattr(self.request, field)

        # to unify all length=0 DocumentArray (or any other results) will simply considered as None
        # otherwise the executor has to handle DocArray(0)
        if len(result):
            return result

    def _get_docs_matrix(self, field) -> List['DocumentArray']:
        """DocumentArray from (multiple) requests

        :param field: either `docs` or `groundtruths`

        .. # noqa: DAR201"""
        if self.expect_parts > 1:
            result = [getattr(r, field) for r in reversed(self.partial_requests)]
        else:
            result = [getattr(self.request, field)]

        # to unify all length=0 DocumentArray (or any other results) will simply considered as None
        # otherwise, the executor has to handle [None, None, None] or [DocArray(0), DocArray(0), DocArray(0)]
        len_r = sum(len(r) for r in result)
        if len_r:
            return result

    @property
    def docs(self) -> 'DocumentArray':
        """Return a DocumentArray by concatenate (multiple) ``requests.docs``

        .. # noqa: DAR201"""
        return self._get_docs('docs')

    @property
    def groundtruths(self) -> 'DocumentArray':
        """Return a DocumentArray by concatenate (multiple) ``requests.groundtruths``

        .. # noqa: DAR201"""
        return self._get_docs('groundtruths')

    @property
    def docs_matrix(self) -> List['DocumentArray']:
        """Return a list of DocumentArray from multiple requests

        .. # noqa: DAR201"""
        return self._get_docs_matrix('docs')

    @property
    def groundtruths_matrix(self) -> List['DocumentArray']:
        """A flattened DocumentArray from (multiple) requests

        .. # noqa: DAR201"""
        return self._get_docs_matrix('groundtruths')

    @property
    def envelope(self) -> 'jina_pb2.EnvelopeProto':
        """Get the current message envelope

        .. # noqa: DAR201
        """
        return self._message.envelope



"""Module for hello world argparser"""
import argparse

from .base import set_base_parser
from .helper import add_arg_group, _SHOW_ALL_ARGS, _chf
from ..helper import random_identity


def mixin_hw_base_parser(parser):
    """Add the arguments for hello world to the parser

    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='General')
    gp.add_argument(
        '--workdir',
        type=str,
        default=random_identity(),
        help='The workdir for hello-world demo'
        'all data, indices, shards and outputs will be saved there',
    )
    gp.add_argument(
        '--download-proxy', type=str, help='The proxy when downloading sample data'
    )


def set_hello_parser(parser=None):
    """
    Set the hello parser

    :param parser: the parser configure
    """

    if not parser:
        parser = set_base_parser()

    spp = parser.add_subparsers(
        dest='hello',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get detailed information about each sub-command',
        required=True,
    )

    set_hw_parser(
        spp.add_parser(
            'fashion',
            help='Start a simple end2end fashion images index & search demo. '
            'This demo requires no extra dependencies.',
            description='Run a fashion search demo',
            formatter_class=_chf,
        )
    )

    set_hw_chatbot_parser(
        spp.add_parser(
            'chatbot',
            help='''
Start a simple Covid-19 chatbot.

Remarks:

- Pytorch, transformers & FastAPI are required to run this demo. To install all dependencies, use

    pip install "jina[chatbot]"

- The indexing could take 1~2 minute on a CPU machine.
''',
            description='Run a chatbot QA demo',
            formatter_class=_chf,
        )
    )

    set_hw_multimodal_parser(
        spp.add_parser(
            'multimodal',
            help='''
Start a simple multimodal document search.

Remarks:

- Pytorch, torchvision, transformers & FastAPI are required to run this demo. To install all dependencies, use

    pip install "jina[multimodal]"

- The indexing could take 2~3 minute on a CPU machine.
- Downloading the dataset could take ~1 minute depending on your network.
''',
            description='Run a multimodal search demo',
            formatter_class=_chf,
        )
    )

    set_hw_fork_parser(
        spp.add_parser(
            'fork',
            help='Fork a hello world project to a local directory, and start to build your own project on it.',
            description='Fork a hello world project to a local directory.',
            formatter_class=_chf,
        )
    )


def set_hw_parser(parser=None):
    """Set the hello world parser

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    gp = add_arg_group(parser, title='Index')
    gp.add_argument(
        '--index-data-url',
        type=str,
        default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz',
        help='The url of index data (should be in idx3-ubyte.gz format)',
    )
    gp.add_argument(
        '--index-labels-url',
        type=str,
        default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz',
        help='The url of index labels data (should be in idx3-ubyte.gz format)',
    )

    gp = add_arg_group(parser, title='Search')
    gp.add_argument(
        '--query-data-url',
        type=str,
        default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz',
        help='The url of query data (should be in idx3-ubyte.gz format)',
    )
    gp.add_argument(
        '--query-labels-url',
        type=str,
        default='http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz',
        help='The url of query labels data (should be in idx3-ubyte.gz format)',
    )

    gp.add_argument(
        '--request-size',
        type=int,
        default=1024,
        help='The request size in indexing (the maximum number of documents that will be included in a '
        'Request before sending it)',
    )

    gp.add_argument(
        '--num-query', type=int, default=128, help='The number of queries to visualize'
    )
    gp.add_argument(
        '--top-k', type=int, default=50, help='Top-k results to retrieve and visualize'
    )
    return parser


def set_hw_chatbot_parser(parser=None):
    """Set the parser for the hello world chatbot

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    parser.add_argument(
        '--index-data-url',
        type=str,
        default='https://static.jina.ai/chatbot/dataset.csv',
        help='The url of index csv data',
    )
    parser.add_argument(
        '--port-expose',
        type=int,
        default=8080,
        help='The port of the host exposed to the public',
    )
    parser.add_argument(
        '--parallel',
        type=int,
        default=2,
        help='The number of parallel when index and query',
    )
    parser.add_argument(
        '--unblock-query-flow',
        action='store_true',
        default=False,
        help='Do not block the query flow' if _SHOW_ALL_ARGS else argparse.SUPPRESS,
    )
    return parser


def set_hw_fork_parser(parser=None):
    """Set the parser for forking hello world demo

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    parser.add_argument(
        'project',
        type=str,
        choices=['fashion', 'chatbot', 'multimodal'],
        help='The hello world project to fork',
    )

    parser.add_argument(
        'destination',
        type=str,
        help='The dest directory of the forked project. Note, it can not be an existing path.',
    )

    return parser


def set_hw_multimodal_parser(parser=None):
    """Set the parser for the hello world multimodal

    :param parser: the parser configure
    :return: the new parser
    """

    if not parser:
        parser = set_base_parser()

    mixin_hw_base_parser(parser)
    parser.add_argument(
        '--index-data-url',
        type=str,
        default='https://static.jina.ai/multimodal/people-img.zip',
        help='The url of index csv data',
    )
    parser.add_argument(
        '--port-expose',
        type=int,
        default=8080,
        help='The port of the host exposed to the public',
    )
    parser.add_argument(
        '--unblock-query-flow',
        action='store_true',
        default=False,
        help='Do not block the query flow' if _SHOW_ALL_ARGS else argparse.SUPPRESS,
    )
    return parser
